# pylint: disable=no-member, unsubscriptable-object

from django.db import models
from django.urls import reverse
from django.utils import timezone

from discordbot.botmodules.bots import VierGewinntBot
from discordbot.botmodules.parser import HTMLCleaner

from discordbot.config import DOMAIN

from asgiref.sync import sync_to_async

import time
import uuid
import re

# Basic

class Member(models.Model):
    server = models.ForeignKey("Server", on_delete=models.CASCADE)
    user = models.ForeignKey("User", on_delete=models.CASCADE, related_name="servers")

    def __str__(self):
        return str(self.server.name)+" - "+str(self.user.name)
    __str__.short_description = "Mitglied"

    class Meta():
        verbose_name = "Mitglied"
        verbose_name_plural = "Mitglieder"

    objects = models.Manager()

class Server(models.Model):
    id = models.CharField("Discord ID", primary_key=True, max_length=20)
    name = models.CharField("Name", default="", blank=True, max_length=100)
    members = models.ManyToManyField("User", through="Member")

    playlist = models.ForeignKey("Playlist", on_delete=models.SET_NULL, default=None, blank=True, null=True, related_name="+")

    @sync_to_async
    def getPlaylist(self):
        if self.playlist is None:
            self.playlist = Playlist.objects.create(server=self, title="Playlist in "+self.name)
            self.save()
        return self.playlist

    def reportCount(self):
        return self.reports.count()
    reportCount.short_description = "Reports"

    @sync_to_async
    def getReports(self, user=None):
        if user is None:
            description = ""
            for member in self.members.all():
                count = int(member.reportCount(server=self))
                if count == 1:
                    description += f"{member.mention} (Ein Report)\n"
                elif count > 1:
                    description += f"{member.mention} ({count} Reports)\n"
            return description
        else:
            return user.getReports(server=self)

    def memberCount(self):
        return self.members.count()
    memberCount.short_description = "Members"

    def __str__(self):
        return self.name+" ("+str(self.id)+")"
    __str__.short_description = "Guild"

    class Meta():
        verbose_name = "Guild"
        verbose_name_plural = "Guilds"

    objects = models.Manager()

class User(models.Model):
    id = models.CharField("Discord ID", primary_key=True, max_length=20)
    name = models.CharField("Name", default="", blank=True, max_length=100)

    def reportCount(self, **filters):
        return self.reports.filter(**filters).count()
    reportCount.short_description = "Reports"

    def createdReportCount(self, **filters):
        return self.created_reports.filter(**filters).count()
    createdReportCount.short_description = "Created reports"

    def serverCount(self, **filters):
        return self.servers.filter(**filters).count()
    serverCount.short_description = "Server count"

    @sync_to_async
    def joinServer(self, server):
        if not server.members.filter(id=self.id).exists():
            server.members.add(self)

    def getReports(self, **filters):
        reports = self.reports.filter(**filters)
        return [
            report.getEmbedField() for report in reports
        ]

    @property
    def mention(self):
        return "<@"+str(self.id)+">"

    def __str__(self):
        return self.name+" ("+str(self.id)+")"
    __str__.short_description = "User"

    class Meta():
        verbose_name = "Benutzer"
        verbose_name_plural = "Benutzer"
    
    objects = models.Manager()

# Music

class AudioSource(models.Model):
    url_source = models.TextField("Url (Source)")

    url_watch = models.URLField("Url (Watch)", default="", blank=True)
    url_thumbnail = models.URLField("Url (Thumbnail)", default="", blank=True)

    title = models.CharField("Title", max_length=256)
    description = models.TextField("Description", max_length=2048)
    duration = models.IntegerField("Duration", default=0, blank=True)

    uploader_name = models.CharField("Uploader name", max_length=256, default="", blank=True)
    uploader_url = models.URLField("Uploader Url", default="", blank=True)

    @classmethod
    @sync_to_async
    def create_from_dict(self, data:dict):
        # if "formats" in data:
        #     data.pop("formats")
        # if "thumbnails" in data:
        #     data.pop("thumbnails")
        # print(data)

        if not data is None:
            url_source = data.get('url', '')

            url_watch = data.get('webpage_url', url_source)[:200]
            url_thumbnail = data.get('thumbnail', "")[:200]

            title = data.get('title', 'Unbekannter Titel')[:256]
            description = data.get('description', "Keine Beschreibung gefunden.")[:2048]
            duration = int(data.get('duration', 0))
            
            uploader_name = data.get('uploader', "")[:256]
            uploader_url = data.get('uploader_url', "")[:200]
            
            return self.objects.create(
                url_source=url_source,

                url_watch=url_watch,
                url_thumbnail=url_thumbnail,
                
                title=title,
                description=description,
                duration=duration,

                uploader_name=uploader_name,
                uploader_url=uploader_url,
            )
        else:
            return None

    @property
    def clickable(self):
        return f"[{self.title}]({self.url_watch}) [{self.duration_calc}]"

    @property
    def duration_calc(self):
        h = str(self.duration//3600)
        m = str(self.duration%3600//60)
        s = str(self.duration%3600%60)
        return f"{h}:{m}:{s}" if h != "0" else f"{m}:{s}"

    def __str__(self):
        return f"{self.title} [{self.duration_calc}]"
    __str__.short_description = "Audio source"

    class Meta():
        verbose_name = "Audio source"
        verbose_name_plural = "Audio sources"

    objects = models.Manager()

class PlaylistPosition(models.Model):
    source = models.ForeignKey('AudioSource', on_delete=models.PROTECT)
    queue = models.ForeignKey('Playlist', on_delete=models.CASCADE)
    position = models.PositiveSmallIntegerField("Position")

    class Meta():
        verbose_name = "Audio queue position"
        verbose_name_plural = "Audio queue positions"

    objects = models.Manager()

class Playlist(models.Model):
    server = models.ForeignKey('Server', on_delete=models.CASCADE, related_name="playlists")
    title = models.CharField("Title", max_length=256, default="Server default")

    sources = models.ManyToManyField('AudioSource', through="PlaylistPosition")
    current_pos = models.PositiveSmallIntegerField("Current position", default=0)

    @sync_to_async
    def addSource(self, source):
        pos = self.sources.count()+1
        self.sources.add(source, through_defaults={"position": pos})
        return pos

    @sync_to_async
    def removePosition(self, pos: int):
        if pos > 0 and pos <= self.sources.count() and not self.current_pos == pos:
            self.sources.through.objects.filter(position=pos).delete()
            for s in self.sources.through.objects.filter(position__gt=pos):
                s.position -= 1
                s.save()

    @sync_to_async
    def switchPositions(self, pos1:int, pos2:int):
        if self.sources.count() > max(pos1, pos2):
            o1 = self.sources.through.objects.get(playlist=self, position=pos1)
            o2 = self.sources.through.objects.get(playlist=self, position=pos2)
            o1.position = pos2
            o2.position = pos1
            o1.save()
            o2.save()
            return True
        return False

    @sync_to_async
    def getAll(self):
        return list(self.sources.all().order_by("position"))

    @sync_to_async
    def hasNext(self):
        return self.sources.through.objects.filter(playlist=self, position__gt=self.current_pos).exists()

    @sync_to_async
    def next(self):
        self.current_pos += 1
        obj = self.sources.through.objects.get(playlist=self, position=self.current_pos)
        self.save()
        return obj

    class Meta():
        verbose_name = "Audio queue"
        verbose_name_plural = "Audio queues"

    objects = models.Manager()


# Support

class Report(models.Model):
    server = models.ForeignKey("Server", on_delete=models.CASCADE, related_name="reports")
    user = models.ForeignKey("User", on_delete=models.CASCADE, related_name="reports")
    reported_by = models.ForeignKey("User", on_delete=models.SET_NULL, null=True, related_name="created_reports", verbose_name="Gemeldet von")

    reason = models.CharField("Grund", default="", blank=True, max_length=250)
    timestamp = models.DateTimeField("Zeitpunkt", auto_now_add=True)

    def getEmbedField(self):
        return (
            f"{self.timestamp.strftime('%Y/%m/%d %H:%M:%S')} ({self.pk})",
            f"{self.reported_by.mention} - {self.reason}",
            False
        )

    def __str__(self):
        return "Report #"+str(self.pk)
    __str__.short_description = "Report"

    class Meta():
        verbose_name = "Report"
        verbose_name_plural = "Reports"

    objects = models.Manager()

# Games

## AmongUs

AMONGUS_PLAYER_COLORS = {
    'red':      ("#C51111", (179, 17, 17),   ":heart:",                 '❤'),
    'blue':     ("#132FD2", (19, 47, 210),   ":blue_heart:",            '💙'),
    'green':    ("#127F2D", (18, 127, 45),   ":green_heart:",           '💚'),
    'pink':     ("#ED53B9", (237, 83, 185),  ":heartpulse:",            '💗'),
    'orange':   ("#EF7D0E", (239, 125, 14),  ":orange_heart:",          '🧡'),
    'yellow':   ("#F3F558", (243, 246, 88),  ":yellow_heart:",          '💛'),
    'black':    ("#3F484E", (63, 72, 78),    ":black_heart:",           '🖤'),
    'white':    ("#D5E0EF", (213, 224, 239), ":white_heart:",           '🤍'),
    'purple':   ("#6B30BC", (107, 48, 188),  ":purple_heart:",          '💜'),
    'brown':    ("#72491E", (114, 37, 30),   ":brown_heart:",           '🤎'),
    'cyan':     ("#39FEDB", (57, 254, 219),  ":regional_indicator_c:",  '🇨'),
    'lime':     ("#50EF3A", (80, 239, 58),   ":regional_indicator_l:",  '🇱'),
}

AMONGUS_EMOJI_COLORS = {v[3]: k for k, v in AMONGUS_PLAYER_COLORS.items()}

class AmongUsGame(models.Model):
    api_key = models.UUIDField("API Key", default=uuid.uuid4)

    creator = models.ForeignKey("User", on_delete=models.CASCADE)
    guild = models.ForeignKey("Server", on_delete=models.CASCADE)
    
    voice_channel_id = models.CharField("Voice channel ID", max_length=20)
    text_channel_id = models.CharField("Text channel ID", max_length=20)
    text_message_id = models.CharField("Text message ID", max_length=20, default=None, null=True)

    code = models.CharField("Gamecode", max_length=6, default="", blank=True)

    last_edited = models.DateTimeField("Last edited", auto_now=True)
    last_tracking_data = models.DateTimeField("Last tracking data", default=None, null=True)

    state_ingame = models.BooleanField("In game", default=False)
    state_meeting = models.BooleanField("Meeting ongoing", default=False)

    p_red_name = models.CharField("Red's name", max_length=50, default="", blank=True)
    p_red_alive = models.BooleanField("Red's alive", default=True)
    p_red_exists = models.BooleanField("Red exists", default=False)
    p_red_userid = models.CharField("Red's Discord ID", max_length=50, default="", blank=True)
    p_blue_name = models.CharField("Blue's name", max_length=50, default="", blank=True)
    p_blue_alive = models.BooleanField("Blue's alive", default=True)
    p_blue_exists = models.BooleanField("Blue exists", default=False)
    p_blue_userid = models.CharField("Blue's Discord ID", max_length=50, default="", blank=True)
    p_green_name = models.CharField("Green's name", max_length=50, default="", blank=True)
    p_green_alive = models.BooleanField("Green's alive", default=True)
    p_green_exists = models.BooleanField("Green exists", default=False)
    p_green_userid = models.CharField("Green's Discord ID", max_length=50, default="", blank=True)
    p_pink_name = models.CharField("Pink's name", max_length=50, default="", blank=True)
    p_pink_alive = models.BooleanField("Pink's alive", default=True)
    p_pink_exists = models.BooleanField("Pink exists", default=False)
    p_pink_userid = models.CharField("Pink's Discord ID", max_length=50, default="", blank=True)
    p_orange_name = models.CharField("Orange's name", max_length=50, default="", blank=True)
    p_orange_alive = models.BooleanField("Orange's alive", default=True)
    p_orange_exists = models.BooleanField("Orange exists", default=False)
    p_orange_userid = models.CharField("Orange's Discord ID", max_length=50, default="", blank=True)
    p_yellow_name = models.CharField("Yellow's name", max_length=50, default="", blank=True)
    p_yellow_alive = models.BooleanField("Yellow's alive", default=True)
    p_yellow_exists = models.BooleanField("Yellow exists", default=False)
    p_yellow_userid = models.CharField("Yellow's Discord ID", max_length=50, default="", blank=True)
    p_black_name = models.CharField("Black's name", max_length=50, default="", blank=True)
    p_black_alive = models.BooleanField("Black's alive", default=True)
    p_black_exists = models.BooleanField("Black exists", default=False)
    p_black_userid = models.CharField("Black's Discord ID", max_length=50, default="", blank=True)
    p_white_name = models.CharField("White's name", max_length=50, default="", blank=True)
    p_white_alive = models.BooleanField("White's alive", default=True)
    p_white_exists = models.BooleanField("White exists", default=False)
    p_white_userid = models.CharField("White's Discord ID", max_length=50, default="", blank=True)
    p_purple_name = models.CharField("Purple's name", max_length=50, default="", blank=True)
    p_purple_alive = models.BooleanField("Purple's alive", default=True)
    p_purple_exists = models.BooleanField("Purple exists", default=False)
    p_purple_userid = models.CharField("Purple's Discord ID", max_length=50, default="", blank=True)
    p_brown_name = models.CharField("Brown's name", max_length=50, default="", blank=True)
    p_brown_alive = models.BooleanField("Brown's alive", default=True)
    p_brown_exists = models.BooleanField("Brown exists", default=False)
    p_brown_userid = models.CharField("Brown's Discord ID", max_length=50, default="", blank=True)
    p_cyan_name = models.CharField("Cyan's name", max_length=50, default="", blank=True)
    p_cyan_alive = models.BooleanField("Cyan's alive", default=True)
    p_cyan_exists = models.BooleanField("Cyan exists", default=False)
    p_cyan_userid = models.CharField("Cyan's Discord ID", max_length=50, default="", blank=True)
    p_lime_name = models.CharField("Lime's name", max_length=50, default="", blank=True)
    p_lime_alive = models.BooleanField("Lime's alive", default=True)
    p_lime_exists = models.BooleanField("Lime exists", default=False)
    p_lime_userid = models.CharField("Lime's Discord ID", max_length=50, default="", blank=True)

    
    def get_tracker_url(self):
        if DOMAIN:
            return DOMAIN+reverse("discordbot:amongus-tracker-post")
        else:
            return "Frage den Bot-Owner!"

    def reset(self):
        self.code = ""
        self.state_ingame = False
        self.state_meeting = False
        for c in AMONGUS_PLAYER_COLORS:
            setattr(self, f'p_{c}_name', "")
            setattr(self, f'p_{c}_exists', False)
            setattr(self, f'p_{c}_alive', True)

    def post_data(self, data:dict):
        if "api_key" in data and data["api_key"] == str(self.api_key):
            if "reset" in data and data["reset"]:
                self.reset()
            if "code" in data:
                self.code = str(data["code"])
            if "state" in data:
                if "ingame" in data["state"]:
                    self.state_ingame = bool(data["state"]["ingame"])
                if "meeting" in data["state"]:
                    self.state_meeting = bool(data["state"]["meeting"])
            if "players" in data:
                for c in data["players"]:
                    if c in AMONGUS_PLAYER_COLORS:
                        if "name" in data["players"][c]:
                            setattr(self, f'p_{c}_name', data["players"][c]["name"])
                        if "alive" in data["players"][c]:
                            setattr(self, f'p_{c}_alive', data["players"][c]["alive"])
                        if "exists" in data["players"][c]:
                            setattr(self, f'p_{c}_exists', data["players"][c]["exists"])
                            
            self.last_tracking_data = timezone.now()
            self.save()
            return {"success": True}
        else:
            return {"error": "invalid-api-key", "error_message": "API Key is missing or wrong!"}

    def get_data(self):
        players = {c: {
                "name": getattr(self, f'p_{c}_name'),
                "alive": getattr(self, f'p_{c}_alive'),
                "userid": getattr(self, f'p_{c}_userid'),
                "exists": getattr(self, f'p_{c}_exists'),
            } for c in AMONGUS_PLAYER_COLORS
        }

        return {
            "id": self.pk,
            "code": self.code,
            "state": {
                "ingame": self.state_ingame,
                "meeting": self.state_meeting,
            },
            "players": players,
            "last_edited": self.last_edited,
            "last_tracking_data": self.last_tracking_data,
        }

    # User

    @sync_to_async
    def remove_user(self, userid):
        for c in AMONGUS_PLAYER_COLORS:
            if getattr(self, f'p_{c}_userid') == str(userid):
                setattr(self, f'p_{c}_userid', "")
        self.save()

    @sync_to_async
    def set_user(self, userid, color):
        for c in AMONGUS_PLAYER_COLORS:
            if getattr(self, f'p_{c}_userid') == str(userid):
                setattr(self, f'p_{c}_userid', "")
        setattr(self, f'p_{color}_userid', str(userid))
        self.save()

    def __str__(self):
        return "AmongUs #"+str(self.pk)

    class Meta:
        verbose_name = 'AmongUs Game'
        verbose_name_plural = 'AmongUs Games'
        unique_together = ('guild', 'creator',)

    objects = models.Manager()


## VierGewinnt

VIERGEWINNT_PLAYER_EMOJIS = ["⬛", "🟥", "🟨"]
VIERGEWINNT_WALL_EMOJI = "🟦"
VIERGEWINNT_NUMBER_EMOJIS = ["1️⃣", "2️⃣", "3️⃣", "4️⃣", "5️⃣", "6️⃣", "7️⃣", "8️⃣", "9️⃣", "🔟"]

def VIERGEWINNT_DEFAULT_GAME():
    return [[0 for _ in range(7)] for _ in range(6)]

class VierGewinntGame(models.Model):
    width = models.PositiveSmallIntegerField("Width", default=7)
    height = models.PositiveSmallIntegerField("Height", default=6)

    game = models.JSONField("Game", default=VIERGEWINNT_DEFAULT_GAME)

    finished = models.BooleanField("Finished", default=False)

    current_player = models.PositiveSmallIntegerField("Current player", default=1)

    player_1_id = models.CharField("Player 1 ID", max_length=32)
    player_2_id = models.CharField("Player 2 ID", max_length=32, default=None, null=True)
    winner_id = models.CharField("Winner ID", max_length=32, default=None, null=True)

    channel_id = models.CharField("Channel ID", max_length=32)
    message_id = models.CharField("Message ID", max_length=32)

    time_created = models.DateTimeField("Time created", auto_now_add=True)
    time_edited = models.DateTimeField("Time edited", auto_now=True)

    # Create

    @classmethod
    def create(self, width:int=7, height:int=6, **kwargs):
        width = 10 if width > 10 else 4 if width < 4 else width
        height = 4 if height < 4 else 14 if height > 14 else height
        kwargs["game"] = [[0 for _ in range(width)] for _ in range(height)]
        return self.objects.create(width=width, height=height, **kwargs)

    # Show

    def _get_gameboard(self):
        W = VIERGEWINNT_WALL_EMOJI
        P = VIERGEWINNT_PLAYER_EMOJIS
        N = VIERGEWINNT_NUMBER_EMOJIS

        numberline = (W+"".join(N[:int(self.width)])+W+"\n")
        gamelines = "\n".join([(W+"".join([P[p] for p in row])+W) for row in self.rows])+"\n"
        
        return numberline + gamelines + numberline

    def _get_players(self):
        pl = []
        for i in range(1, 2+1):
            pid = getattr(self, f"player_{ i }_id")
            if pid is None:
                pl.append(VIERGEWINNT_PLAYER_EMOJIS[i]+" BOT")
            else:
                pl.append(VIERGEWINNT_PLAYER_EMOJIS[i]+" <@"+pid+">")
        return "\n".join(pl)

    def _get_game_info(self):
        if self.finished:
            if self.winner_id:
                if self.winner_id == "BOT":
                    return "Das Spiel ist beendet! Du hast leider verloren!"
                else:
                    return f"Das Spiel ist beendet! Gewonnen hat <@{self.winner_id}>!"
            return "Das Spiel ist beendet! Unentschieden!"
        else:
            p = getattr(self, f"player_{ self.current_player }_id")
            if p is None:
                return f"{ VIERGEWINNT_PLAYER_EMOJIS[self.current_player] } Ich bin an der Reihe! (Berchne einen guten Zug...)"
            else:
                return f"{ VIERGEWINNT_PLAYER_EMOJIS[self.current_player] } <@{ p }> ist an der Reihe!"

    def get_description(self):
        return self._get_game_info()+"\n\n"+self._get_players()+"\n\n"+self._get_gameboard()
    get_description.short_description = "Game"

    # Get

    @property
    def rows(self):
        return self.game

    @property
    def cols(self):
        return [[self.rows[h][w] for h in range(self.height)] for w in range(self.width)]

    @property
    def dias(self):
        dias = []
        w = 0 
        h = self.height-1
        while w < self.width:
            dia = []
            _w, _h = w, h

            while _w < self.width and _h < self.height:
                dia.append(self.rows[_h][_w])
                _w += 1
                _h += 1

            dias.append(dia)
            if h > 0:
                h -= 1
            else:
                w += 1
        w = 0
        h = 0
        while w < self.width:
            dia = []
            _w, _h = w, h

            while _w < self.width and _h >= 0:
                dia.append(self.rows[_h][_w])
                _w += 1
                _h -= 1

            dias.append(dia)
            if h < self.height-1:
                h += 1
            else:
                w += 1
        return dias

    # Functions

    def _next_player(self):
        self.current_player = (self.current_player+1 if self.current_player < 2 else 1)

    def _can_add_to_column(self, n: int):
        if n >= self.width:
            return None
        return not self.rows[0][n]

    def _add_to_column(self, n:int):
        if self._can_add_to_column(n):
            for h in range(self.height-1, -1, -1):
                if not self.game[h][n]:
                    self.rows[h][n] = self.current_player
                    self._next_player()
                    return True
        else:
            return False

    # Checks

    def _is_full(self):
        return not 0 in self.rows[0]

    def _get_winner(self):
        for ls in [self.rows, self.cols, self.dias]:
            for l in ls:
                for i in range(len(l)-3):
                    if (not l[i] == 0) and (l[i] == l[i+1]) and (l[i+1] == l[i+2]) and (l[i+2] == l[i+3]):
                        return l[i]
        return 0

    # Process Input

    def process(self, col:int, playerid):
        if not (self.finished or self.winner_id):
            if getattr(self, "player_"+str(self.current_player)+"_id", None) == str(playerid):
                if self._add_to_column(col):
                    w = self._get_winner()
                    if w:
                        self.winner_id = getattr(self, "player_"+str(w)+"_id", None)
                        self.finished = True
                    elif self._is_full():
                        self.finished = True
                    return True
        return False

    def process_bot(self):
        if not self.finished and getattr(self, f"player_{self.current_player}_id") is None: 
            if self._add_to_column(VierGewinntBot.get_best_move(board=self.game, botnr=self.current_player, maxdepth=4)):
                w = self._get_winner()
                if w:
                    self.winner_id = "BOT"
                    self.finished = True
                elif self._is_full():
                    self.finished = True
                return True
        return False

    # Meta

    class Meta:
        verbose_name = 'VierGewinnt Game'
        verbose_name_plural = 'VierGewinnt Games'

    objects = models.Manager()

# Support

## BotPermission (for Remote)

class BotPermission(models.Model):
    id_1 = models.CharField("ID 1", max_length=32)
    id_2 = models.CharField("ID 2", max_length=32)
    typ = models.CharField("Typ", max_length=32)

    objects = models.Manager()

# Notifier

NOTIFIER_WHERE_TYPES = [
    ("channel", "Kanal"),
    ("member",  "Mitglied (DM)"),
]

NOTIFIER_FREQUENCIES = [
    ("hour",    "Stündlich"),
    ("minute",  "Minütlich"),
]

class NotifierSub(models.Model):
    name = models.CharField("Name", default="Unbenannte Mitteilung", max_length=50)

    where_type = models.CharField("Wohin: Typ", max_length=8, choices=NOTIFIER_WHERE_TYPES)
    where_id = models.CharField("Wohin: ID", max_length=32)

    frequency = models.CharField("Häufigkeit", max_length=8, choices=NOTIFIER_FREQUENCIES)

    url = models.URLField("URL")

    must_contain_regex = models.CharField("Muss Regex enthalten", default="", blank=True, max_length=32)

    last_hash = models.CharField("Letzter Hash", max_length=32, editable=False)

    def process(self):
        data = HTMLCleaner.from_url(self.url)
        thishash = str(hash(data))
        if thishash != str(self.last_hash):
            self.last_hash = thishash
            if (not self.must_contain_regex) or re.search(self.must_contain_regex, data):
                return data
            return True
        return False

    def __str__(self):
        return self.name
    __str__.short_description = "NotifierSub"

    class Meta:
        verbose_name = 'NotifierSub'
        verbose_name_plural = 'NotifierSubs'

    objects = models.Manager()



from discordbot.botevents.on_command_error import setup as setup_on_command_error
from discordbot.botevents.on_command import setup as setup_on_command
from discordbot.botevents.on_message import setup as setup_on_message
from discordbot.botevents.general import setup as setup_general

def setup(bot):
    setup_on_command_error(bot)
    setup_on_command(bot)
    setup_on_message(bot)
    setup_general(bot)


from rich import print
from tqdm import tqdm

class VierGewinntBot():
    # Constants
    
    ## Ratings [Sequence], (Points for P1, P2)

    RATING = [
        ([1, 1, 1, 1],      ( 100000000000, -100000000000)),
        ([2, 2, 2, 2],      (-100000000000,  100000000000)), 
        ([0, 1, 1, 1, 0],   ( 1000,         -1000000)),
        ([0, 2, 2, 2, 0],   (-1000000,       1000)), 
        ([0, 1, 1, 1],      ( 100,          -1000000)), 
        ([0, 2, 2, 2],      (-1000000,       100)), 
        ([1, 1, 1, 0],      ( 100,          -1000000)),
        ([2, 2, 2, 0],      (-1000000,       100)),
        ([1, 0, 1, 1],      ( 100,          -1000000)), 
        ([2, 0, 2, 2],      (-1000000,       100)), 
        ([1, 1, 0, 1],      ( 100,          -1000000)), 
        ([2, 2, 0, 2],      (-1000000,       100)), 
        ([0, 1, 1, 0],      ( 10,           -10)), 
        ([0, 2, 2, 0],      (-10,            10)), 
        ([0, 1, 0],         ( 1,            -1)),
        ([0, 2, 0],         (-1,             1)),
    ]

    FINISHED_CRITERIUM = [
        [1, 1, 1, 1],
        [2, 2, 2, 2],
    ]

    # Utils

    @classmethod
    def copyboard(self, board):
        return [l.copy() for l in board]

    @classmethod
    def seq_in_list(self, sequence, listtotest):
        if len(sequence) <= len(listtotest):
            return ', '.join(map(str, sequence)) in ', '.join(map(str, listtotest))
        else:
            return False

    @classmethod
    def get_lines(self, board):
        width, height = len(board[0]), len(board)

        # Horizontal lines
        lines = self.copyboard(board)

        # Vertical lines
        lines += [[board[h][w] for h in range(height)] for w in range(width)]

        # Diagonal lines
        w = 0
        h = height-1
        while w < width:
            dia = []
            _w, _h = w, h

            while _w < width and _h < height:
                dia.append(board[_h][_w])
                _w += 1
                _h += 1

            lines.append(dia)
            if h > 0:
                h -= 1
            else:
                w += 1
        w = 0
        h = 0
        while w < width:
            dia = []
            _w, _h = w, h

            while _w < width and _h >= 0:
                dia.append(board[_h][_w])
                _w += 1
                _h -= 1

            lines.append(dia)
            if h < height-1:
                h += 1
            else:
                w += 1
        return lines

    @classmethod
    def get_rating(self, board, playernr):
        lines = self.get_lines(board)
        score = 0
        for sequence, points in self.RATING:
            for line in lines:
                if self.seq_in_list(sequence, line):
                    if playernr == 1:
                        score += points[0]
                    elif playernr == 2:
                        score += points[1]
        return score

    @classmethod
    def is_finished(self, board):
        lines = self.get_lines(board)
        for seq in self.FINISHED_CRITERIUM:
            for line in lines:
                if self.seq_in_list(seq, line):
                    return True
        return False

    # Preview move

    @classmethod
    def get_move_preview(self, board, position, playernr):
        board = self.copyboard(board)
        for i in range(len(board)-1, -1, -1):
            if board[i][position] == 0:
                board[i][position] = playernr
                return board

    # Recursive

    @classmethod
    def _get_best_move(self, board, playernr=2, level=0, alpha=-100000000000, beta=100000000000, botnr=2, maxdepth=4):
        scores = []
        for i in range(len(board[0])):
            if board[0][i] == 0:
                if level < maxdepth:
                    if not self.is_finished(board):
                        otherplayer = 2 if playernr == 1 else 1

                        score = self._get_best_move(
                            self.get_move_preview(board, i, playernr), 
                            playernr=otherplayer, 
                            level=level+1, 
                            alpha=alpha, 
                            beta=beta, 
                            botnr=botnr, 
                            maxdepth=maxdepth
                        )

                        scores.append(score)

                        alpha = max(alpha, self.get_rating(board, otherplayer))
                        if beta <= alpha:
                            break
                    else:
                        return self.get_rating(board, botnr)
                else:
                    return self.get_rating(board, botnr)
            else:
                return self.get_rating(board, botnr)

        if playernr == botnr:
            return max(scores)
        else:
            return min(scores)

    # Main

    @classmethod
    def get_best_move(self, board, botnr=2, maxdepth=4):
        scores = []
        for i in tqdm(range(len(board[0])), desc="[VierGewinntBot]"):
            if board[0][i] == 0:
                otherplayer = 2 if botnr == 1 else 1

                score = self._get_best_move(
                    self.get_move_preview(board, i, botnr), 
                    playernr=otherplayer, 
                    level=1, 
                    alpha=-100000000000, 
                    beta=100000000000, 
                    botnr=botnr, 
                    maxdepth=maxdepth
                )

                scores.append(score)
            else:
                scores.append(-100000000000)

        bestmove = scores.index(max(scores))

        print(f"[VierGewinntBot] - Result ({ bestmove }):", scores)
        return bestmove



from discord.ext import commands
from discord import utils
import typing
import socket
import dns

#from discordbot.config import 
from discordbot.errors import ErrorMessage

class Domains(commands.Cog):
    def __init__(self, bot):
        self.bot = bot
        self.color = 0x5a03fc

    @commands.command(
        brief="Erhalte die IP einer Domain",
        description="Frage die IP-Adresse ab, welche hinter einer Domain steckt.",
        aliases=["ip"],
        help="",
        usage="<Domain>"
        )
    async def getip(self, ctx, domain:str):
        try:
            ip = socket.gethostbyname(domain)
            await ctx.sendEmbed(
                title="IP-Info",
                description=f"Die IP hinter '{domain}' lautet '{ip}'!",
            )
        except socket.gaierror:
            raise ErrorMessage(f"Die Domain '{domain}' konnte nicht gefunden werden!")

    @commands.command(
        brief="Frage DNS-Einträge ab",
        description="Frage bestimmte DNS-Einträge einer Domain ab.",
        aliases=["getdns"],
        help="",
        usage="<Domain> <Typ (CNAME, A, MX)>"
    )
    async def dns(self, ctx, domain: str, typ: str):
        typ = typ.upper()
        try:
            result = dns.resolver.query(domain, typ)
            if typ == "A":
                await ctx.sendEmbed(
                    title="DNS-Info",
                    description=f"DNS-Einträge des Typs A für '{domain}'!",
                    inline=False,
                    fields=[
                        ("IP", ipval.to_text()) for ipval in result
                    ]
                )
            elif typ == "CNAME":
                await ctx.sendEmbed(
                    title="DNS-Info",
                    description=f"DNS-Einträge des Typs CNAME für '{domain}'!",
                    inline=False,
                    fields=[
                        ("CNAME Target", cnameval.target) for cnameval in result
                    ]
                )
            elif typ == "MX":
                await ctx.sendEmbed(
                    title="DNS-Info",
                    description=f"DNS-Einträge des Typs MX für '{domain}'!",
                    inline=False,
                    fields=[
                        ("MX Record", mxdata.exchange) for mxdata in result
                    ]
                )
            else:
                await ctx.sendEmbed(
                    title="DNS-Info",
                    description=f"DNS-Einträge des Typs '{typ}' für '{domain}'!",
                    inline=False,
                    fields=[
                        ("Eintrag", str(data)) for data in result
                    ]
                )
        except dns.resolver.NXDOMAIN:
            raise ErrorMessage(
                f"Die Domain '{domain}' konnte nicht gefunden werden!")
        except dns.resolver.NoAnswer:
            raise ErrorMessage(
                f"Für die Domain '{domain}' konnten keine DNS-Einträge des Typs '{typ}' gefunden werden!")
        except dns.rdatatype.UnknownRdatatype:
            raise ErrorMessage(
                f"Unbekannter DNS-Record Typ: {typ}")

def setup(bot):
    bot.add_cog(Domains(bot))



# pylint: disable=no-member

from discord.ext import commands, tasks
from discord import Embed, User, TextChannel, utils
from datetime import datetime as d
import typing

from discordbot.botmodules.serverdata import DjangoConnection
from discordbot.models import NotifierSub

class Notifier(commands.Cog):
    def __init__(self, bot):
        self.bot = bot
        self.color = 0xeb8f34

        print('[Notifier Background Tasks] - Started!')
        self.notifier_backgroundtasks_minute.start()
        self.notifier_backgroundtasks_hour.start()

    def cog_unload(self):
        print("[Notifier Background Tasks] - Stopped!")
        self.notifier_backgroundtasks_minute.cancel()
        self.notifier_backgroundtasks_hour.cancel()

    #

    async def notifier_update(self, frequency:str, send:bool=False):
        for sub in await DjangoConnection._list(NotifierSub, frequency=frequency):
            data = sub.process()
            if data:
                await DjangoConnection._save(sub)
            if send and isinstance(data, str):
                if sub.where_type == "channel":
                    where = self.bot.get_channel(int(sub.where_id)) or await self.bot.fetch_channel(int(sub.where_id))
                elif sub.where_type == "member":
                    where = self.bot.get_user(int(sub.where_id)) or await self.bot.fetch_user(int(sub.where_id))
                emb = self.bot.getEmbed(
                    title="Neue Änderung!",
                    description=str(data),
                    authorname=sub.name,
                    authorurl=sub.url,
                    color=self.color,
                )
                await where.send(embed=emb)

    #

    @tasks.loop(minutes=1)
    async def notifier_backgroundtasks_minute(self):
        await self.notifier_update("minute", send=True)

    @notifier_backgroundtasks_minute.before_loop
    async def notifier_backgroundtasks_minute_before(self):
        await self.notifier_update("minute", send=False)

    @tasks.loop(hours=1)
    async def notifier_backgroundtasks_hour(self):
        await self.notifier_update("hour", send=True)

    @notifier_backgroundtasks_hour.before_loop
    async def notifier_backgroundtasks_hour_before(self):
        await self.notifier_update("hour", send=False)

    #

    # @commands.group(
    #     brief="Erhalte Notifications zu Webseiten",
    #     description='Erhalte eine Nachricht, wenn sich auf einer Webseite etwas ändert.',
    #     aliases=['notify', 'noti'],
    #     usage=""
    # )
    # async def notifier(self, ctx):
    #     if ctx.invoked_subcommand is None:
    #         await ctx.send_help()

    

def setup(bot):
    bot.add_cog(Notifier(bot))



from discord.ext import commands
from discord import Embed, Member, User, Webhook, utils

from discordbot.models import BotPermission, Report
from discordbot.errors import ErrorMessage

import time
import typing

class Support(commands.Cog):
    def __init__(self, bot):
        self.bot = bot
        self.color = 0xffdf00

    # Reports

    async def send_report(self, ctx, report):
        try:
            channel = utils.get(ctx.guild.channels, name="reports")
            await ctx.sendEmbed(
                title=f"Neuer Report ({report.pk})",
                fields=[
                    ("Gemeldeter Benutzer", report.user.mention),
                    ("Gemeldet von", report.reported_by.mention),
                    ("Grund", report.reason),
                    ("Zeitpunkt", report.timestamp.strftime('%Y/%m/%d %H:%M:%S')),
                ],
                inline=False,
                receiver=channel,
                footerurl="",
                footertext="",
                )
        except Exception as e:
            print("[Support] - No #reports channel found!", e)

    @commands.group(
        brief="Melde Spieler und sehe dir Meldungen an",
        description="Dieses Modul unterstützt dich dabei, Personen, welche sich falsch verhalten, auf diesem Server zu erkennen",
        usage="<Unterbefehl> <Argumente>",
        aliases=["rp", "rep", "report"],
    )
    @commands.guild_only()
    async def reports(self, ctx):
        if ctx.invoked_subcommand is None:
            await ctx.send_help()

    @reports.command(
        name="create",
        brief='Melde einen Spieler',
        description='Benutze diesen Command um Spieler zu melden',
        aliases=["add", "+"],
        usage="<Member> [Grund]"
        )
    async def reports_create(self, ctx, member: Member, *args):
        Grund = " ".join(args)
        Grund = Grund if Grund.rstrip(" ") else "Leer"
        report = await ctx.database.createReport(dc_user=member, reason=Grund)
        await ctx.sendEmbed(title="Benutzer Gemeldet", fields=[("Betroffener",member.mention),("Grund",Grund)])
        await self.send_report(ctx, report)
        return

    @reports.command(
        name="delete",
        brief='Lösche einen Report',
        description='Dieser Command ist für alle mit einer der folgenden Rollen verfügbar: Moderator, Supporter, Administrator und Admin',
        aliases=["del", "-"],
        usage="<ID>"
    )
    @commands.has_any_role("Moderator", "Supporter", "Admin", "Administrator")
    async def reports_delete(self, ctx, *repids):
        for repid in repids:
            if await ctx.database.deleteReport(repid=repid):
                await ctx.sendEmbed(title=f"Report #{repid} gelöscht!")
            else:
                raise ErrorMessage(f"Report mit der ID {repid} wurde nicht gefunden!")
        return

    @reports.command(
        name="view",
        brief='Siehe Reports an',
        description='Benutze diesen Command um alle Reports zu sehen',
        aliases=["get", "show", "list"],
        help="Mit /reports view kannst du dir alle Reports ansehen, mit /reports view [Member] kannst du dir Reports eines bestimmten Mitglieds ansehen.",
        usage="[Member]"
        )
    @commands.has_any_role("Moderator", "Supporter", "Admin", "Administrator")
    async def reports_view(self, ctx, member:Member=None):
        if member == None:
            await ctx.sendEmbed(
                title="Server Reports",
                description=(
                    f"Reports auf {ctx.guild.name}\n\n"+
                    await ctx.database.getReports()
                )
            )
        else:
            await ctx.sendEmbed(
                title="Mitglieder Reports",
                description=("Mitglied: "+member.mention),
                fields=await ctx.database.getReports(dc_user=member),
            )
        return

    # Remote
    
    @commands.group(
        brief="Lasse jemanden Commands für dich ausführen",
        description="Überlasse jemandem das Recht, Commands als dich auszuführen.",
        usage="<Unterbefehl> <Argumente>",
    )
    async def remote(self, ctx):
        if ctx.invoked_subcommand is None:
            await ctx.send_help()

    @remote.command(
        name="allow",
        brief="Füge einen erlaubten Spieler hinzu",
        aliases=["+", "add"],
        usage="<Member>",
        help="Dieser Befehl kann ziemlich viel Mist machen, vorallem wenn du Rechte hast! Also pass auf!",
    )
    async def remote_allow(self, ctx, member: typing.Union[Member, User]):
        if not await ctx.database._has(BotPermission, id_1=str(ctx.author.id), id_2=str(member.id), typ="remote_permission"):
            await ctx.database._create(BotPermission, id_1=str(ctx.author.id), id_2=str(member.id), typ="remote_permission")
            await ctx.sendEmbed(
                title="Benutzer erlaubt",
                description=member.mention+" darf nun Befehle für dich ausführen!"
            )
        else:
            raise commands.BadArgument("Dieser Benutzer darf bereits Befehle für dich ausführen!")

    @remote.command(
        name="allowraw",
        brief="Füge eine erlaubte ID hinzu",
        aliases=["allowwebhook", "++", "addraw"],
        usage="<ID>",
        help="Dies wird gebraucht, um Webhooks die Erlaubnis zu geben, Commands für dich auszuführen. Dieser Befehl kann ziemlich viel Mist machen, vorallem wenn du Rechte hast! Also pass auf!",
        hidden=True,
    )
    async def remote_allowraw(self, ctx, id: int):
        if not await ctx.database._has(BotPermission, id_1=str(ctx.author.id), id_2=str(id), typ="remote_permission"):
            await ctx.database._create(BotPermission, id_1=str(ctx.author.id), id_2=str(id), typ="remote_permission")
            await ctx.sendEmbed(
                title="ID erlaubt",
                description=str(id)+" darf nun Befehle für dich ausführen!"
            )
        else:
            raise commands.BadArgument("Diese ID darf bereits Befehle für dich ausführen!")

    @remote.command(
        name="disallow",
        brief="Entferne einen erlaubten Spieler",
        aliases=["-", "remove"],
        usage="<Member>",
    )
    async def remote_disallow(self, ctx, member: typing.Union[Member, User]):
        if await ctx.database._has(BotPermission, id_1=str(ctx.author.id), id_2=str(member.id), typ="remote_permission"):
            await ctx.database._listdelete(BotPermission, id_1=str(ctx.author.id), id_2=str(member.id), typ="remote_permission")
            await ctx.sendEmbed(
                title="Benutzer verboten", 
                description=member.mention+" darf nun nicht mehr Befehle für dich ausführen!"
            )
        else:
            raise commands.BadArgument("Dieser Benutzer steht nicht auf deiner Liste!")

    @remote.command(
        name="disallowraw",
        brief="Entferne eine erlaubte ID",
        aliases=["--", "removeraw"],
        usage="<Member>",
        help="Dies wird gebraucht, um Webhooks die Erlaubnis zu geben, Commands für dich auszuführen.",
        hidden=True,
    )
    async def remote_disallowraw(self, ctx, id: int):
        if await ctx.database._has(BotPermission, id_1=str(ctx.author.id), id_2=str(id), typ="remote_permission"):
            await ctx.database._listdelete(BotPermission, id_1=str(ctx.author.id), id_2=str(id), typ="remote_permission")
            await ctx.sendEmbed(
                title="ID verboten",
                description=str(id)+" darf nun nicht mehr Befehle für dich ausführen!"
            )
        else:
            raise commands.BadArgument(
                "Diese ID steht nicht auf deiner Liste!")

    @remote.command(
        name="list",
        brief="Liste die erlaubten Spieler auf",
        aliases=[],
        usage=""
    )
    async def remote_list(self, ctx):
        perms = await ctx.database._list(BotPermission, id_1=str(ctx.author.id), typ="remote_permission")
        await ctx.sendEmbed(
            title="Benutzerliste",
            description="Folgende Benutzer/IDs dürfen Befehle für dich ausführen:\n\n"+"\n".join([f"<@{i.id_2}> ({i.id_2})" for i in perms])
        )

    @remote.command(
        name="run",
        brief="Führe einen Befehl als jemand anderes aus",
        aliases=[],
        usage="<Member> <Command> [Argumente]",
        help="Falls Unterbefehle verwendet werden, benutze bitte befehl_unterbefehl als Command",
    )
    async def remote_run(self, ctx, member: typing.Union[Member, User], command: str, *args):
        if await ctx.database._has(BotPermission, id_1=str(member.id), id_2=str(ctx.author.id), typ="remote_permission") or await ctx.database._has(BotPermission, id_1=str(member.id), id_2=str(ctx.message.webhook_id), typ="remote_permission"):
            await ctx.invoke_as(member, command, *args)
        else:
            raise commands.BadArgument("Du darfs keine Befehle als diesen Benutzer ausführen!   ")

def setup(bot):
    bot.add_cog(Support(bot))



import os
from django.conf import settings

# Discordbot Configuration
## Helper function

def _s(name:str, default=None, prefix:str="DISCORDBOT_"):
    return getattr(settings, prefix+name, os.getenv(prefix+name, default))

## Config

DEBUG = _s("DEBUG", False)

DOMAIN = _s("DOMAIN", None)

## Discordbot Config

ALLOW_BOTS = _s("ALLOW_BOTS", False)
ALLOW_WEBHOOKS = _s("ALLOW_WEBHOOKS", True)

ALL_PREFIXES = _s("ALL_PREFIXES", ["/", "!", "$", ".", "-", ">", "?"])
MAIN_PREFIXES = _s("MAIN_PREFIXES", ["/"])

EXTENSIONFOLDER = "discordbot.botcmds"
EXTENSIONS = _s("EXTENSIONS", ['basic','support','moderation','games','help','channels','music','owneronly','converters','embedgenerator','notifier', 'polls', 'domains'])

## Basic

INVITE_OWNER = "https://rebrand.ly/RUdiscord"
INVITE_BOT = "https://rebrand.ly/RUdiscordbot"

SPAM_ALLOWED = False

## Help

HELP_HIDDEN_COGS = ["owneronly", "notifier"]

## Music

MUSIC_MODULE = _s("MUSIC_MODULE", False)

FILESPATH = _s("FILESPATH", os.path.join(os.path.dirname(os.path.realpath(__file__)), "botfiles"))
MEMESPATH = _s("MEMESPATH", os.path.join(FILESPATH, "memes"))
FFMPEGPATH = _s("FFMPEGPATH", os.path.join(FILESPATH, "ffmpeg.exe"))

FFMPEG_OPTIONS = {
    'options': '-vn',
    'executable': FFMPEGPATH,
}

FFMPEG_OPTIONS_STREAM = {
    'options': '-vn',
    'before_options': " -reconnect 1 -reconnect_streamed 1 -reconnect_delay_max 5",
    'executable': FFMPEGPATH,
}

RADIOS = {
    "swisspop": "http://www.radioswisspop.ch/live/mp3.m3u",
    "nrjbern":  "https://energybern.ice.infomaniak.ch/energybern-high.mp3",
}



# Generated by Django 3.1.3 on 2020-11-10 17:26

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('discordbot', '0013_notifiersub'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='notifiersub',
            options={'verbose_name': 'NotifierSub', 'verbose_name_plural': 'NotifierSubs'},
        ),
        migrations.AlterField(
            model_name='notifiersub',
            name='last_hash',
            field=models.CharField(editable=False, max_length=32, verbose_name='Letzter Hash'),
        ),
        migrations.AlterField(
            model_name='notifiersub',
            name='must_contain_regex',
            field=models.CharField(blank=True, default='', max_length=32, verbose_name='Muss Regex enthalten'),
        ),
        migrations.AlterField(
            model_name='notifiersub',
            name='where_id',
            field=models.CharField(max_length=32, verbose_name='ID'),
        ),
    ]



# pylint: disable=unused-variable

from discord.ext import commands
from discord import Embed

from discordbot.config import DEBUG
from discordbot.errors import ErrorMessage

DEBUG_NO_RAISE_FOR = [
    commands.CommandNotFound, 
    commands.MissingRequiredArgument, 
    commands.BadArgument, 
    commands.CommandOnCooldown,
    commands.MissingAnyRole,
    commands.MissingRole,
]

def setup(bot):
    @bot.event
    async def on_command_error(ctx, error):
        if isinstance(error, ErrorMessage):
            await ctx.sendEmbed(title="Fehler", color=0xff0000, description=str(error), fields=[("Nachricht", ctx.message.content)])
        else:
            EMBED = ctx.getEmbed(title="Fehler", color=0xff0000)
            if isinstance(error, commands.BadArgument):
                EMBED.add_field(name="Beschreibung", value="Du hast ungültige Argumente angegeben!")
            elif isinstance(error, commands.MissingRequiredArgument):
                EMBED.add_field(name="Beschreibung", value="Du hast ein benötigtes Argument weggelassen!")
            elif isinstance(error, commands.CommandNotFound):
                EMBED.add_field(name="Beschreibung", value="Dieser Befehl wurde nicht gefunden!")
            elif isinstance(error, commands.CommandOnCooldown):
                EMBED.add_field(name="Beschreibung", value="Warte, bis du diesen Befehl erneut benutzen kannst!")
            elif isinstance(error, commands.DisabledCommand):
                EMBED.add_field(name="Beschreibung", value="Dieser Command ist aktuell deaktiviert!")
            elif isinstance(error, commands.TooManyArguments):
                EMBED.add_field(name="Beschreibung", value="Du hast zu viele Argumente angegeben!")
            elif isinstance(error, commands.MissingPermissions):
                EMBED.add_field(name="Beschreibung", value="Du hast nicht die nötigen Berechtigungen für diesen Command!")
            elif isinstance(error, commands.BotMissingPermissions):
                EMBED.add_field(name="Beschreibung", value="Ich habe nicht die nötigen Berechtigungen für diesen Command!")
            elif isinstance(error, commands.NoPrivateMessage):
                EMBED.add_field(name="Beschreibung", value="Dieser Command kann nur auf einem Server verwendet werden!")
            elif isinstance(error, commands.PrivateMessageOnly):
                EMBED.add_field(name="Beschreibung", value="Dieser Command kann nur in Direktnachrichten werden!")
            elif isinstance(error, commands.MissingRole):
                EMBED.add_field(name="Beschreibung", value="Du hast die benötigten Rollen nicht, um diesen Befehl auszuführen!")
            elif isinstance(error, commands.MissingAnyRole):
                EMBED.add_field(name="Beschreibung", value="Du hast die benötigten Rollen nicht, um diesen Befehl auszuführen!")
            elif isinstance(error, commands.NotOwner):
                EMBED.add_field(name="Beschreibung", value="Du bist nicht Besitzer des Bots!")
                return # keine Nachricht senden!
            elif isinstance(error, commands.CommandError):
                EMBED.add_field(name="Beschreibung", value="Bei einem Befehl ist ein Fehler aufgetreten!")
            else:
                EMBED.add_field(name="Beschreibung", value="Es ist ein unbekannter Fehler aufgetreten! Vermutlich liegt er nicht bei dir, also melde ihn am besten einem Admin.")
                if DEBUG:
                    print("[Command] - Bei '"+ctx.message.content+"' von '"+ctx.message.author.name+"#"+ctx.message.author.discriminator+"' ist ein Fehler aufgetreten: "+str(error))
            
            if not error == "":
                if "OperationalError: (2006, 'MySQL server has gone away')" in str(error):
                    EMBED.add_field(name="Info", value="Leider konnte die Verbindung zur Datenbank nicht hergestellt werden. Bitte versuche es später noch einmal!")
                else:
                    EMBED.add_field(name="Error", value=str(error) if len(str(error)) < 1024 else str(error)[-1024:-1])
            EMBED.add_field(name="Nachricht", value=ctx.message.content, inline=False)
            await ctx.send(embed=EMBED)

            if DEBUG and not True in [isinstance(error, cls) for cls in DEBUG_NO_RAISE_FOR]:
                raise error



from django.shortcuts import render
from django.http import HttpResponse, JsonResponse
from django.views.decorators.http import require_POST
from django.views.decorators.csrf import csrf_exempt

from discordbot.models import AmongUsGame

import json

# Create your views here.

@require_POST
@csrf_exempt
def amongus_tracker_post(request):
    try:
        data = dict(json.loads(json.loads(request.body)))
        id = int(data["id"])

        if AmongUsGame.objects.filter(id=id).exists():
            game = AmongUsGame.objects.get(id=id)
            return JsonResponse(game.post_data(data))
        else:
            return JsonResponse({"error": "id-not-found", "error_message": "Game with this ID not found!"})
    except Exception as e:
        return JsonResponse({"error": "invalid-data", "error_message": str(e)})



from discord.ext import commands
from discord import Embed
import random

from discordbot.config import HELP_HIDDEN_COGS

class Help(commands.Cog):
    def __init__(self, bot):
        self.bot = bot
        self.color = 0x000000

    @commands.command(
        name='help',
        brief='Erhalte Hilfe zu Commands',
        aliases=["hilfe", "commands", "command", "?", "cmds"],
        help="Benutze diesen Command für genauere Hilfe zur jeweiligen Kategorie / zum jeweiligen Befehl/Unterbefehl",
        usage="[Kategorie/Befehl/Befehl Unterbefehl]"
    )
    async def help(self, ctx, search:str='*', subsearch:str=None):
        description = ["Syntax: `/help [Kategorie/Befehl/Befehl Unterbefehl]`\n\n"]
        fields = []

        def addCog(cog, include_subcommands=False, hide_cogs=HELP_HIDDEN_COGS):
            if not cog.qualified_name.lower() in hide_cogs:
                commands_list = ''
                for cmd in cog.walk_commands():
                    if include_subcommands:
                        depth = 0
                        currentcmd = cmd
                        while currentcmd.parent:
                            depth += 1
                            currentcmd = currentcmd.parent 

                        if depth == 0:
                            commands_list += f'\n**{cmd.name}** - {cmd.brief}\n'
                        elif depth == 1:
                            commands_list += f'- {cmd.name} '
                        # else:
                        #    commands_list += "- "*depth + f'**{cmd.name}**\n'
                    elif not cmd.parent:
                        commands_list += f'**{cmd.name}** - {cmd.brief}\n'

                fields.append((cog.qualified_name, commands_list.replace("\n\n", "\n")+'\u200b'))

        def addGroup(grp):
            commands_list = ''
            for cmd in grp.commands:
                if not cmd.hidden:
                    commands_list += f'**{cmd.name}** - {cmd.brief}\n'

            fields.append(("Unterbefehle von "+grp.qualified_name, commands_list+'\u200b'))

        def addCommand(cmd):
            if cmd.hidden:
                description[0] += "\n*Dies ist ein versteckter Befehl*\n\n"
            if isinstance(cmd, commands.Group):
                if subsearch:
                    subcmds = dict((alias, cmd) for cmd in cmd.commands for alias in cmd.aliases+[cmd.name.lower()])
                    if subsearch in subcmds:
                        addCommand(subcmds[subsearch])
                        return
                addGroup(cmd)

            cmdname = cmd.name
            currentcmd = cmd
            while currentcmd.parent:
                currentcmd = currentcmd.parent
                cmdname = currentcmd.name + " " + cmdname

            cmdbrief = " - "+cmd.brief if cmd.brief else ""
            help_text = f'```/{cmdname}{cmdbrief}```\n' 

            if cmd.description:
                help_text += f'Beschrieb: `{cmd.description}`\n\n'
            if cmd.help:
                help_text += f'Hilfe: `{cmd.help}`\n\n'
            if cmd.aliases:
                help_text += f'Alias: `{"`, `".join(cmd.aliases)}`'

            help_text += f'\nSyntax: `/{cmdname}{" "+cmd.usage if cmd.usage is not None else ""}`\n\u200b'
            description[0] += help_text
    

        cogs = dict((k.lower(), v) for k, v in dict(self.bot.cogs).items())
        cmds = dict((alias, cmd) for cmd in self.bot.commands for alias in cmd.aliases+[cmd.name.lower()])

        if search == '*':
            for cog in cogs:
                addCog(cogs[cog])

        elif search in cmds:
            addCommand(cmds[search])

        elif search in cogs:
            addCog(cogs[search], include_subcommands=True, hide_cogs=[])

        else:
            raise commands.BadArgument("Ungültige(r) Kategorie/Befehl.\nBenutze den `/help` Befehl um alle Kategorien und Befehle zu sehen.")

        await ctx.sendEmbed(
            title='Hilfe',
            inline=False,
            description=description[0],
            fields=fields,
        )


def setup(bot):
    bot.add_cog(Help(bot))



# Generated by Django 3.0.4 on 2020-07-01 14:55

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('discordbot', '0001_initial'),
    ]

    operations = [
        migrations.RenameField(
            model_name='server',
            old_name='discordid',
            new_name='id',
        ),
        migrations.RenameField(
            model_name='user',
            old_name='discordid',
            new_name='id',
        ),
    ]



# Generated by Django 3.1.3 on 2020-11-13 19:31

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('discordbot', '0015_notifiersub_name'),
    ]

    operations = [
        migrations.AlterField(
            model_name='notifiersub',
            name='where_id',
            field=models.CharField(max_length=32, verbose_name='Wohin: ID'),
        ),
        migrations.AlterField(
            model_name='notifiersub',
            name='where_type',
            field=models.CharField(choices=[('channel', 'Kanal'), ('member', 'Mitglied (DM)')], max_length=8, verbose_name='Wohin: Typ'),
        ),
    ]



from html.parser import HTMLParser as HTMLParserOriginal
import re
import requests


class HTMLCleaner(HTMLParserOriginal):
    def __init__(self, data, convert_charrefs=True):
        super().__init__(convert_charrefs=convert_charrefs)
        self.data = ""

        self.feed(data)

        self.data = "\n".join(i.strip(" ") for i in self.data.split("\n"))

        for i in ["\n\r", "\r\n", "\n\n\n\n"]:
            while i in self.data:
                self.data = self.data.replace(i, "\n\n")

        self.data = self.data.strip(" ")

    def handle_data(self, data):
        self.data += data+"\n"

    def handle_starttag(self, tag, attrs=None):
        if tag in ["del"]:
            self.data += "~~**"
        elif tag in ["li"]:
            self.data += "- "
        elif tag in ["b", "strong"]:
            self.data += "**"
        elif tag in ["i", "em"]:
            self.data += "_"
        elif tag in ["strike"]:
            self.data += "~~"
        elif tag in ["ins"]:
            self.data += "__"
        elif tag in ["hr"]:
            self.data += "\n---\n"

    def handle_endtag(self, tag):
        if tag in ["del"]:
            self.data += "**~~"
        elif tag in ["b", "strong"]:
            self.data += "**"
        elif tag in ["i", "em"]:
            self.data += "_"
        elif tag in ["strike"]:
            self.data += "~~"
        elif tag in ["ins"]:
            self.data += "__"

    @classmethod
    def from_data(self, data):
        return self(data).data

    @classmethod
    def from_url(self, url):
        html = requests.get(url).text
        try:
            data = ">".join(html.split("<body")[1].split(">")[1:]).split("</body>")[0]
        except:
            data = html
        return self(data).data




from discord.ext import commands

from discordbot.botmodules.audio import AudioManager, YouTubePlayer
from discordbot.errors import ErrorMessage

from asgiref.sync import sync_to_async

#####

class MusicQueue():
    def __init__(self, server):
        self.server = server

        self._players = []

    def addPlayer(self, player):
        self._players.append(player)

    def hasPlayer(self):
        return bool(self._players)

    def playNext(self, ctx):
        if self.hasPlayer() and ctx.voice_client and ctx.voice_client.is_connected():
            player = self._players.pop(0)
            player.play(ctx)
            return player
        else:
            return None

    async def sendNowPlaying(self, ctx):
        if ctx.voice_client and ctx.voice_client.source:
            if isinstance(ctx.voice_client.source, YouTubePlayer):
                await ctx.voice_client.source.send(ctx, status="Wird aktuell gespielt.")
        else:
            raise ErrorMessage("Aktuell wird nichts abgespielt.")

    async def sendQueue(self, ctx):
        description = "\n".join(i.getinfo() for i in self._players)
        await ctx.sendEmbed(
            title="Warteschlange",
            color=ctx.command.cog.color,
            description=description,
        )

    async def createYoutubePlayer(self, search, loop=None, stream=False):
        return await YouTubePlayer.from_url(search, queue=self, loop=loop, stream=stream)



class Server():
    _all = {}

    def __init__(self,id):
        self.id = id
        self.musicqueue = MusicQueue(server=self)
        #self.polls = {}

    @classmethod
    def getServer(self, serverid:int):
        if not serverid in self._all:
            self._all[serverid] = Server(serverid)
        return self._all[serverid]




### NEW

from discordbot.models import Server as DB_Server, User as DB_User, Report as DB_Report, Member as DB_Member, AudioSource, Playlist, AmongUsGame, VierGewinntGame, BotPermission, NotifierSub
from django.db import connection, connections

class DjangoConnection():
    def __init__(self, dc_user, dc_guild):
        self.dc_user = dc_user
        self.dc_guild = dc_guild
        self._db_user = None
        self._db_server = None
        self._db_playlist = None
        
    @classmethod
    def ensure_connection(self):
        if connection.connection and not connection.is_usable():
            del connections._connections.default

    # Basic Methods

    @classmethod
    @sync_to_async
    def _save(self, obj):
        self.ensure_connection()
        obj.save()

    @classmethod
    @sync_to_async
    def _delete(self, obj):
        self.ensure_connection()
        obj.delete()

    @classmethod
    async def _del(self, obj):
        await self._delete(obj)

    @classmethod
    @sync_to_async
    def _create(self, model, **kwargs):
        self.ensure_connection()
        return model.objects.create(**kwargs)

    @classmethod
    @sync_to_async
    def _exists(self, model, **filters):
        self.ensure_connection()
        return model.objects.filter(**filters).exists()

    @classmethod
    async def _has(self, model, **filters):
        return await self._exists(model, **filters)

    @classmethod
    @sync_to_async
    def _get(self, model, **filters):
        self.ensure_connection()
        return model.objects.get(**filters)

    @classmethod
    @sync_to_async
    def _list(self, model, _order_by=("pk",), **filters):
        self.ensure_connection()
        return list(model.objects.filter(**filters).order_by(*_order_by))

    @classmethod
    @sync_to_async
    def _listdelete(self, model, **filters):
        self.ensure_connection()
        model.objects.filter(**filters).delete()

    @classmethod
    async def _listdel(self, model, **filters):
        await self._listdelete(model, **filters)

    @classmethod
    async def _getdel(self, model, **filters):
        if await self._has(model, **filters):
            obj = await self._get(DB_Report, **filters)
            await self._delete(obj)
            return True
        return False

    # General

    @classmethod
    async def fetch_user(self, dc_user):
        if not await self._has(DB_User, id=str(dc_user.id)):
            user = await self._create(DB_User, id=str(dc_user.id), name=dc_user.name+"#"+dc_user.discriminator)
        else:
            user = await self._get(DB_User, id=str(dc_user.id))
            if not user.name == (dc_user.name+"#"+dc_user.discriminator):
                user.name = (dc_user.name+"#"+dc_user.discriminator)
                await self._save(user)
        return user

    async def get_user(self):
        if self._db_user is None:
            self._db_user = await self.fetch_user(self.dc_user)
        return self._db_user

    @classmethod
    async def fetch_server(self, dc_guild):
        if not await self._has(DB_Server, id=str(dc_guild.id)):
            server = await self._create(DB_Server, id=str(dc_guild.id), name=dc_guild.name)
        else:
            server = await self._get(DB_Server, id=str(dc_guild.id))
            if not server.name == dc_guild.name:
                server.name = dc_guild.name
                await self._save(server)
        return server

    async def get_server(self):
        if self._db_server is None:
            self._db_server = await self.fetch_server(self.dc_guild)
        return self._db_server

    async def get_playlist(self):
        if self._db_playlist is None:
            server = await self.get_server()
            self._db_playlist = await server.getPlaylist()
        return self._db_playlist

    # Music

    @classmethod
    async def getOrCreateAudioSourceFromDict(self, data):
        if await self._exists(AudioSource, url_watch=data.get("webpage_url", data.get("url", None))):
            audio = await self._get(AudioSource, url_watch=data.get("webpage_url", data.get("url")))
            audio.url_source = data.get("url", "")
            await self._save(audio)
            return audio
        else:
            return await AudioSource.create_from_dict(data)

    # Reports

    async def createReport(self, dc_user, reason:str="", reportedby_dc_user=None):
        server = await self.get_server()
        user = (await self.get_user()) if reportedby_dc_user is None else (await self.fetch_user(reportedby_dc_user))
        await user.joinServer(server)
        reporteduser = await self.fetch_user(dc_user)
        await reporteduser.joinServer(server)
        return await self._create(DB_Report, server=server, user=reporteduser, reported_by=user, reason=reason)

    async def getReports(self, dc_user=None):
        server = await self.get_server()
        if dc_user is None:
            reports = await server.getReports()
            return reports
        else:
            user = await self.fetch_user(dc_user)
            reports = await server.getReports(user=user)
            return reports

    async def deleteReport(self, repid:int):
        server = await self.get_server()
        return await self._getdel(DB_Report, server=server, id=repid)

    # AmongUs

    async def getAmongUsGame(self, **kwargs):
        user = await self.get_user()
        server = await self.get_server()
        return await self._get(AmongUsGame, creator=user, guild=server, **kwargs)

    async def hasAmongUsGame(self, **kwargs):
        user = await self.get_user()
        server = await self.get_server()
        return await self._has(AmongUsGame, creator=user, guild=server, **kwargs)

    async def createAmongUsGame(self, **kwargs):
        user = await self.get_user()
        server = await self.get_server()
        return await self._create(AmongUsGame, creator=user, guild=server, **kwargs)
